/*
  downvfchar - download each outline data of a char from a TrueType font
                as PostScript outline data.
  Copyright (C) 1997 Kazuaki Nakayama (nakayama@math.shinshu-u.ac.jp)
*/
#ifdef VFLIB
#include "dvips.h"
#include "protos.h"
extern char *nextstring;

void adjxy (unsigned int x, unsigned int y,
	unsigned int *nx, unsigned int *ny, int jiscc)
{
  unsigned int tmp;

  switch (jiscc) {
  case 0x2122: case 0x2123:
    /* shift to upper right */
    x += 5000;
    y -= 5100;
    break;
  case 0x213c: case 0x2141:
    /* mirror image */
    tmp = x;
    x = y;
    y = tmp;
    break;
  case 0x2131: case 0x2132: case 0x213d: case 0x213e: case 0x2142:
  case 0x2143: case 0x2144: case 0x2145: case 0x214a: case 0x214b:
  case 0x214c: case 0x214d: case 0x214e: case 0x214f: case 0x2150:
  case 0x2151: case 0x2152: case 0x2153: case 0x2154: case 0x2155:
  case 0x2156: case 0x2157: case 0x2158: case 0x2159: case 0x215a:
  case 0x215b: case 0x222e:
    /* rotate -90 deg */
    tmp = x;
    x = 8192 - y;
    y = tmp;
    break;
  case 0x2421: case 0x2423: case 0x2425: case 0x2427: case 0x2429:
  case 0x2443: case 0x2463: case 0x2465: case 0x2467: case 0x246e:
  case 0x2521: case 0x2523: case 0x2525: case 0x2527: case 0x2529:
  case 0x2543: case 0x2563: case 0x2565: case 0x2567: case 0x256e:
  case 0x2575: case 0x2576:
    /* shift to upper right */
    x += 1000;
    y -= 1000;
    break;
  default:
    break;
  }
  *nx = x; *ny = y;
}

void sendxy (unsigned int x, unsigned int y, int jiscc)
{
  unsigned int tmpx, tmpy;

  adjxy(x, y, &tmpx, &tmpy, jiscc);
  numout((integer)tmpx);
  numout((integer)tmpy);
}

void sendarcpts (double x, double y,
	double r, double arg1, double arg2)
{
  (void)sprintf(nextstring, "%g", x);
  cmdout(nextstring);
  (void)sprintf(nextstring, "%g", y);
  cmdout(nextstring);
  (void)sprintf(nextstring, "%g", r);
  cmdout(nextstring);
  (void)sprintf(nextstring, "%g", arg1);
  cmdout(nextstring);
  (void)sprintf(nextstring, "%g", arg2);
  cmdout(nextstring);
}

#ifndef M_PI
#define M_PI 3.1415926535897932384626
#endif
/* Borrowed from the djgpp patch version 1.3.  See README for details. */
#define radians_to_degrees (180.0/M_PI)
double atan360 (double y, double x)
{
  double ang;

  if((ang = atan2(y, x) * radians_to_degrees) < 0)
    ang += 360.0;
  return ang;
}

void sendarc (unsigned int x1, unsigned int y1, 
		 unsigned int x2, unsigned int y2,
		 unsigned int x3, unsigned int y3,
		 int jiscc, int dir)
{
  unsigned int tmpx1, tmpy1, tmpx2, tmpy2, tmpx3, tmpy3;
  int dx1, dy1, dx3, dy3, z;
  double r, cx, cy, arg, arg1, arg2, arg3;


  if (dir) {
    adjxy(x1, y1, &tmpx1, &tmpy1, jiscc);
    adjxy(x2, y2, &tmpx2, &tmpy2, jiscc);
    adjxy(x3, y3, &tmpx3, &tmpy3, jiscc);
  } else {
    tmpx1 = x1, tmpy1 = y1;
    tmpx1 = x2, tmpy1 = y2;
    tmpx1 = x3, tmpy1 = y3;
  }

  dx1 = tmpx1 - tmpx2; dy1 = tmpy1 - tmpy2;
  dx3 = tmpx3 - tmpx2; dy3 = tmpy3 - tmpy2;

  if ((z = dx1*dy3 - dy1*dx3) == 0) {
    if ((dx1 == dx3) && (dy1 == dy3)) {
      cx = (double)dx1/2.0;
      cy = (double)dy1/2.0;
      r = sqrt(cx*cx + cy*cy);
      arg = atan360((double)dy1 - cy, (double)dx1 - cx);
      sendarcpts(cx + (double)tmpx2, cy + (double)tmpy2,
		 r, arg, arg+(double)360);
      cmdout("arc");
      return;
    } else {
      numout((integer)tmpx1);
      numout((integer)tmpy1);
      cmdout("lineto");
      numout((integer)tmpx3);
      numout((integer)tmpy3);
      cmdout("lineto");
      return;
    }
  }

  cx = (double)((dx1*dx1 + dy1*dy1)*dy3 - (dx3*dx3 + dy3*dy3)*dy1) / z / 2.0;
  cy = - (double)((dx1*dx1 + dy1*dy1)*dx3 - (dx3*dx3 + dy3*dy3)*dx1) / z / 2.0;
  r = sqrt(cx*cx + cy*cy);

  arg1 = atan360((double)dy1 - cy, (double)dx1 - cx);
  arg2 = atan360(-cy, -cx);
  arg3 = atan360((double)dy3 - cy, (double)dx3 - cx);

  if(arg1 == arg3) {
    sendarcpts(cx + (double)tmpx2, cy + (double)tmpy2, r, 0.0, 360.0);
    cmdout("arc");
    return;
  } else if(arg1 < arg3) {
    if(arg1 <= arg2 && arg2 <= arg3) {
      sendarcpts(cx + (double)tmpx2, cy + (double)tmpy2, r, arg1, arg3);
      cmdout("arc");
      return;
    } else {
      sendarcpts(cx + (double)tmpx2, cy + (double)tmpy2, r, arg1, arg3);
      cmdout("arcn");
      return;
    }
  } else {
    if(arg3 <= arg2 && arg2 <= arg1) {
      sendarcpts(cx + (double)tmpx2, cy + (double)tmpy2, r, arg1, arg3);
      cmdout("arcn");
      return;
    } else {
      sendarcpts(cx + (double)tmpx2, cy + (double)tmpy2, r, arg1, arg3);
      cmdout("arc");
      return;
    }
  }
}

#define GETX(x) (VFD_GET_X(x)-OUTLINE_OFFSET)
#define GETY(y) (VFD_GET_Y(y)-OUTLINE_OFFSET)
#define ISTOKEN(x)  (((x)&VFD_TOKEN)==VFD_TOKEN)
#define ISCHAR(x)   (((x)&VFD_CHAR)==VFD_CHAR)
#define ISC1(x)     (((x)&VFD_CWCURV)==VFD_CWCURV)
#define ISC2(x)     (((x)&VFD_CCWCURV)==VFD_CCWCURV)
#define ISCURVE(x)  (ISC1(x)||ISC2(x))
#define ISLINE(x)   (((x)&VFD_LINE)==VFD_LINE)
#define ISARC(x)    (((x)&VFD_ARC)==VFD_ARC)
#define ISBEZIER(x) (((x)&VFD_BEZ)==VFD_BEZ)
#define ISEND(x)    ((x)==0L)
#define VFD_OBJS    (VFD_LINE | VFD_ARC | VFD_BEZ)
void sendvfdata (long *vfdata, int jiscc, integer dir)
{
  long cmd, next, *ptr;
  unsigned int  ix, iy, x1, y1, x2, y2, x3, y3, tmpx, tmpy;

  if (vfdata == NULL) {
    extern char errbuf[];
    (void)sprintf(errbuf, "! sendvfdata: No outline dada.");
    error(errbuf);
  }
  ptr = &vfdata[2];
  cmdout("newpath");
  while (!ISEND(*ptr)) {
    if (!ISCURVE(*ptr)) {     /* Should be "begin closed path". */
      extern char errbuf[];
      (void)sprintf(errbuf, "! sendvfdata: Wrong outline data.");
      error(errbuf);
    }
    cmd = *ptr++;
    ix = GETX(*ptr);
    iy = GETY(*ptr);
    ptr++;
    if (dir) {
      sendxy(ix, iy, jiscc);
    } else {
      numout((integer)ix);
      numout((integer)iy);
    }
    cmdout("moveto");
    if (ISARC(cmd)) {
      x3 = ix;
      y3 = iy;
    }
    while (!ISCURVE(*ptr)&&!ISEND(*ptr)) {
      switch (cmd&VFD_OBJS) {
      case VFD_LINE:
	break;
      case VFD_BEZ:
	/* Can pick up at least two points */
	x1 = GETX(*ptr);
	y1 = GETY(*ptr);
	ptr++;
	x2 = GETX(*ptr);
	y2 = GETY(*ptr);
	ptr++;
	break;
      case VFD_ARC:
	/* ??? Never happen? */
	x1 = x3;
	y1 = y3;
	x2 = GETX(*ptr);
	y2 = GETY(*ptr);
	ptr++;
	break;
      default:
	{
	  extern char errbuf[];
	  (void)sprintf(errbuf, "! sendvfdata: error1.");
	  error(errbuf);
	}
      }
      if (ISTOKEN(*ptr)||ISEND(*ptr)) {
	if (ISCHAR(*ptr)||ISEND(*ptr)) {
	  x3 = ix;
	  y3 = iy;
	} else {
	  next = *ptr++;
	  x3 = GETX(*ptr);
	  y3 = GETY(*ptr);
	  ptr++;
	}
      } else {
	next = cmd;
	x3 = GETX(*ptr);
	y3 = GETY(*ptr);
	ptr++;
      }
      switch (cmd&VFD_OBJS) {
      case VFD_LINE:
	if (dir) {
	  sendxy(x3, y3, jiscc);
	} else {
	  numout((integer)x3);
	  numout((integer)y3);
	}
	cmdout("lineto");
	break;
      case VFD_BEZ:
	if (dir) {
	  sendxy(x1, y1, jiscc);
	  sendxy(x2, y2, jiscc);
	  sendxy(x3, y3, jiscc);
	} else {
	  numout((integer)x1);
	  numout((integer)y1);
	  numout((integer)x2);
	  numout((integer)y2);
	  numout((integer)x3);
	  numout((integer)y3);
	}
	cmdout("curveto");
	break;
      case VFD_ARC:
	/* ??? Never happen? */
	sendarc(x1, y1, x2, y2, x3, y3, jiscc, dir);
	break;
      default:
	{
	  extern char errbuf[];
	  (void)sprintf(errbuf, "! sendvfdata: error2.");
	  error(errbuf);
	}
      }
      cmd = next;
    }
    cmdout("closepath");
  }
  cmdout("eofill");
/*  cmdout("stroke"); */
}
/* End of borrowed part. */

void downvfchar (int fd, int jiscc, integer dir)
{
  long *vfdata;

  vfdata = VF_GetOutline(jiscc, fd);
  if (vfdata == NULL) {
    extern char errbuf[];
    (void)sprintf(errbuf,
                  "! Failed to get bitmap data for char 0x%02x.", jiscc);
    error(errbuf);
  }
  specialout('{');
  sendvfdata(vfdata, jiscc, dir);
  VF_FreeOutline(vfdata, fd);
  specialout('}');

  numout((integer)(jiscc & 0xff));
  cmdout("VD");
}
#endif
