/*
 *  KANJI Code conversion routines.
 */

#ifndef PTEXENC_PTEXENC_H
#define PTEXENC_PTEXENC_H

#include <sys/types.h>
#include <ptexenc/ptxconfig.h>
#include <ptexenc/protos.h>
#include <stdio.h>

extern PTXDLL char *ptexenc_version_string;

#define KANJI_OPTS "{jis|euc|sjis|utf8}"

extern PTXDLL string get_enc_string P1H(void);
extern PTXDLL int    set_enc_string P2H(string file, string inter);
#define getencstring  get_enc_string
#define setencstring  set_enc_string

/* decide if internal Kanji encode is SJIS/UPTEX or not */
extern PTXDLL int  is_internalSJIS P1H(void);
extern PTXDLL int  is_internalUPTEX P1H(void);
#define isinternalSJIS  is_internalSJIS
#define isinternalUPTEX is_internalUPTEX

/* check char range */
extern PTXDLL int ismultichr P3H(int length, int nth, int c);
extern PTXDLL int iskanji1 P1H(int c);
extern PTXDLL int iskanji2 P1H(int c);

/* internal (EUC/SJIS/UPTEX) from/to buffer (EUC/SJIS/UTF-8) code conversion */
extern PTXDLL int multistrlen P3H(unsigned char *s, int len, int pos);
extern PTXDLL long fromBUFF P3H(unsigned char *s, int len, int pos);
extern PTXDLL long toBUFF P1H(long inter);

/* internal (EUC/SJIS/UPTEX) from/to DVI (JIS/UCS) code conversion */
#define fromDVI fromJIS
#define toDVI   toJIS

/* JIS/EUC/SJIS/KUTN/UCS to internal (EUC/SJIS/UPTEX) code conversion */
/* (only for \euc primitive, etc.) */
extern PTXDLL long toJIS P1H(long kcode);
extern PTXDLL long fromJIS P1H(long jis);
extern PTXDLL long fromEUC P1H(long euc);
extern PTXDLL long fromSJIS P1H(long sjis);
extern PTXDLL long fromKUTEN P1H(long kuten);
extern PTXDLL long fromUCS P1H(long ucs);

/* fputs/putc with encoding conversion */
extern PTXDLL int putc2 P2H(int c, FILE* fp);
extern PTXDLL int fputs2 P2H(const char* s, FILE* fp);

extern PTXDLL long input_line2 P5H(FILE* fp, string buff, long pos,
                                    const long buffsize, int* lastchar);
#define inputline2(fp,buff,pos,size) input_line2(fp,buff,pos,size,NULL)
extern PTXDLL long input_line3 P4H(FILE* fp, string buff, long pos,
                                    const long buffsize);
#define inputline3(fp,buff,pos,size) input_line3(fp,buff,pos,size)
/* open/close through nkf */

#ifndef WIN32
extern PTXDLL void nkf_disable P1H(void);
extern PTXDLL FILE *nkf_open P2H(const char* path, const char* mode);
extern PTXDLL int nkf_close P1H(FILE* fp);
#endif /* !WIN32 */

#endif /* PTEXENC_PTEXENC_H */
