/* find-suffix.c: return the stuff after a dot.

   Copyright 1992, 1993, 1995, 2008 Karl Berry.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with this library; if not, see <http://www.gnu.org/licenses/>.  */

#include <kpathsea/config.h>

#include <kpathsea/c-pathch.h>

#ifdef WIN32
#include <kpathsea/knj.h>
#endif

/* Return pointer to first character after `.' in last directory element
   of NAME.  If the name is `foo' or `/foo.bar/baz', we have no extension.  */

string
find_suffix (string name)
{
#ifdef WIN32
  char *p;
  string dot_pos;
  const_string slash_pos;
  for(p=name; *p; p++) {
    if(isknj(*p)) {
      if(*(p+1)) p++;
      continue;
    }
    if(*p == '\\') *p = '/';
  }
  dot_pos = strrchr (name, '.');
#else
  const_string slash_pos;
  string dot_pos = strrchr (name, '.');
#endif

  if (dot_pos == NULL)
    return NULL;
  
  for (slash_pos = name + strlen (name);
#ifdef WIN32
       slash_pos > dot_pos && (*slash_pos != '/');
#else
       slash_pos > dot_pos && !IS_DIR_SEP (*slash_pos);
#endif
       slash_pos--)
    ;
  
  return slash_pos > dot_pos ? NULL : dot_pos + 1;
}



