% jMetaPost change file for MetaPost
%
% written by Michio Matsuyama <fwhw5892@mb.infoweb.ne.jp>
%            Hideyuki Suzuki <hideyuki@sat.t.u-tokyo.ac.jp>
%
% $Id: jmp.ch,v 1.42 2000/03/20 02:55:50 hideyuki Exp $

%
%
% change file of mp.w for jMetaPost (CWEB version)
% Akira Kakuto (translated the WEB version)
%
%

@x
@d default_banner "This is MetaPost, Version 1.208" /* printed when \MP\ starts */
@y
@d default_banner "This is upMetaPost, Version 1.208-0.04-u0.28" /* printed when \MP\ starts */
@z

@x
#define metapost_version "1.208"
@y
#define metapost_version "1.208-0.04-u0.28"
@z

@x
#include <unistd.h> /* for access() */
#endif
@y
#include <unistd.h> /* for access() */
#endif
#include <ptexenc.h>
#include <errno.h>
@z

@x
mp->buf_size = 200;
@y
mp->buf_size = atoi(kpse_var_value("buf_size"));
@z

@x
static boolean mp_input_ln (MP mp, void *f ) {
  /* inputs the next line or returns |false| */
  char *s;
  size_t size = 0; 
  mp->last=mp->first; /* cf.\ Matthew 19\thinspace:\thinspace30 */
  s = (mp->read_ascii_file)(mp,f, &size);
  if (s==NULL)
	return false;
  if (size>0) {
    mp->last = mp->first+size;
    if ( mp->last>=mp->max_buf_stack ) { 
      mp->max_buf_stack=mp->last+1;
      while ( mp->max_buf_stack>mp->buf_size ) {
        mp_reallocate_buffer(mp,(mp->buf_size+(mp->buf_size>>2)));
      }
    }
    (void)memcpy((mp->buffer+mp->first),s,size);
  } 
  free(s);
  return true;
}
@y
static boolean mp_input_ln (MP mp, void *f ) {
  int i = EOF;
  mp->last = input_line2((FILE *)f, mp->buffer, mp->first, mp->buf_size, &i);
  if (i == EOF && errno != EINTR && mp->last == mp->first)
    return false;
  if (i != EOF && i != '\n' && i != '\r') {
    fprintf (stderr, "! Unable to read an entire line---bufsize=%u.\n",
                     (unsigned) mp->buf_size);
    fputs ("Please increase buf_size in texmf.cnf.\n", stderr);
    exit (1);
  }
  if (i == '\r') {
    while ((i = getc (f)) == EOF && errno == EINTR)
      ;
    if (i != '\n')
      ungetc (i, f);
  }
  return true;
}
@z

@x
wterm (mp->banner);
@y
wterm (mp->banner);
wterm(" (");
wterm((char *)getencstring());
wterm(")");
@z

%
% tategaki support
%
% Suppose h==(0,height), d==(0,-depth) && w==(width,0) in horizontal string,
% && h==(height,0), d==(-depth,0) && w==(0,-width) in vertical string.
% Four vertices of the bounding box is h, d, h+w && d+w && those of the
% transformed boundig box is Th, Td, T(h+w) && T(d+w), so that the values
% of Th, Td && Tw are compared here.
@x
@ The height width and depth information stored in a text node determines a
rectangle that needs to be transformed according to the transformation
parameters stored in the text node.

@<Other cases for updating the bounding box...@>=
case mp_text_code: 
  x1=mp_take_scaled(mp, txx_val(p),width_val(p));
  y0=mp_take_scaled(mp, txy_val(p),-depth_val(p));
  y1=mp_take_scaled(mp, txy_val(p),height_val(p));
@y
@ The height width and depth information stored in a text node determines a
rectangle that needs to be transformed according to the transformation
parameters stored in the text node.

Boundig box depends on JFM font ID.

@d yoko_jfm_id   11 /* `yoko-kumi' fonts */
@d tate_jfm_id   9  /* `tate-kumi' fonts */
@d font_jfm_p(A) (mp->font_id[(A)]!=0)

@<Other cases for updating the bounding box...@>=
case mp_text_code:
  if ( mp->font_id[mp_font_n(p)]!=tate_jfm_id ) {
    x1=mp_take_scaled(mp, txx_val(p),width_val(p));
    y0=mp_take_scaled(mp, txy_val(p),-depth_val(p));
    y1=mp_take_scaled(mp, txy_val(p),height_val(p));
  } else {
    x1=mp_take_scaled(mp, txy_val(p),-width_val(p));
    y0=mp_take_scaled(mp, txx_val(p),-depth_val(p));
    y1=mp_take_scaled(mp, txx_val(p),height_val(p));
  };
@z
@x
  x1=mp_take_scaled(mp, tyx_val(p),width_val(p));
  y0=mp_take_scaled(mp, tyy_val(p),-depth_val(p));
  y1=mp_take_scaled(mp, tyy_val(p),height_val(p));
@y
  if ( mp->font_id[mp_font_n(p)]!=tate_jfm_id ) {
    x1=mp_take_scaled(mp, tyx_val(p),width_val(p));
    y0=mp_take_scaled(mp, tyy_val(p),-depth_val(p));
    y1=mp_take_scaled(mp, tyy_val(p),height_val(p));
  } else {
    x1=mp_take_scaled(mp, tyy_val(p),-width_val(p));
    y0=mp_take_scaled(mp, tyx_val(p),-depth_val(p));
    y1=mp_take_scaled(mp, tyx_val(p),height_val(p));
  };
@z

@x
  wlog(mp->banner);
@y
  wlog(mp->banner);
  wlog(" (");
  wlog((char *)(getencstring()));
  wlog(")");
@z

%
% char type pointers
@x
eight_bits  *font_bc;
eight_bits  *font_ec;  /* first and last character code */
@y
eight_bits  *font_bc;
eight_bits  *font_ec;  /* first and last character code */
halfword    *font_nt;
halfword    *font_id;
@z

@x
int         *char_base;  /* base address for |char_info| */
@y
int         *char_base;  /* base address for |char_info| */
int         *ctype_base;
@z

@x
xfree(mp->char_base);
@y
xfree(mp->font_id);
xfree(mp->font_nt);
xfree(mp->char_base);
xfree(mp->ctype_base);
@z

@x
  XREALLOC(mp->char_base,          l, int);
@y
  XREALLOC(mp->font_id,            l, halfword);
  XREALLOC(mp->font_nt,            l, halfword);
  XREALLOC(mp->char_base,          l, int);
  XREALLOC(mp->ctype_base,         l, int);
@z


@x
mp->char_base[null_font]=0;
@y
mp->font_id[null_font]=0;
mp->font_nt[null_font]=0;
mp->char_base[null_font]=0;
mp->ctype_base[null_font]=0;
@z

@x
@d char_mp_info(A,B) mp->font_info[mp->char_base[(A)]+(B)].qqqq
@y
@d char_mp_info(A,B) mp->font_info[mp->char_base[(A)]+(B)].qqqq
@d ctype_char_end(A) (A)].hh.lh
@d ctype_char(A) mp->font_info[mp->ctype_base[(A)]+ctype_char_end
@d ctype_type_end(A) (A)].hh.rh
@d ctype_type(A) mp->font_info[mp->ctype_base[(A)]+ctype_type_end
@z

%
% lookup character type table
@x
void mp_set_text_box (MP mp,pointer p) {
  font_number f; /* |mp_font_n(p)| */
@y
@<Declare JFM function for text measuring@>;
void mp_set_text_box (MP mp,pointer p) {
  int cx; /* code for Japanese two byte character */
  font_number f; /* |mp_font_n(p)| */
@z

@x
  if ( (mp->str_pool[k]<bc)||(mp->str_pool[k]>ec) ) {
    mp_lost_warning(mp, f,k);
  } else { 
    cc=char_mp_info(f,mp->str_pool[k]);
@y
  if ( ((mp->str_pool[k]<bc)||(mp->str_pool[k]>ec)) && (mp->font_id[f]==0) ) {
    mp_lost_warning(mp, f,k);
  } else if (mp->font_id[f]==0) {
      cc=char_mp_info(f,mp->str_pool[k]);
  } else {
    if (multistrlen(mp->str_pool, kk, k)>1) {
      cx=fromBUFF(mp->str_pool, kk, k);
      k=k+multistrlen(mp->str_pool, kk, k)-1;
    } else {
      cx=mp->str_pool[k];
    }
    cc=char_mp_info(f,mp_lookup_ctype(mp, f, cx));

@z

@x
@* \[47] Debugging.
@y
@ @<Declare JFM function for text measuring@>=
int mp_lookup_ctype (MP mp,font_number f, integer c)
{
  int l, u, r, ch;
  l=0; u=mp->font_nt[f]-1;
  while ( l<u ) {
    r=(l+u)/2;
    ch=ctype_char(f)(r);
    if ( (ch==c) ) {
      return ctype_type(f)(r);};
    if ( (ch<c) ) l=r+1;
    else u=r-1;
  };
  return 0;
}
@* \[47] Debugging.
@z
