
%x STRING BCOMMENT INCLUDE TEXCOMMAND TEXINCOMMENT CPCOMMENT
%s CPLUSPLUS

%{
int hfillon = 0;

#define KEY	printf ("{\\%s %s}", keyword_font, yytext); hfillon = 1
#define CPP	printf ("{\\%s \\%s}", cpp_font, yytext); hfillon = 1
#define SYM(x)	printf ("$\\%s$", x); hfillon = 1
#define OUT(x)	printf ("%s", x); hfillon = 1
#define SUB(x)	substitute(x)
#define IND	indent(yytext); hfillon = 0

#include <stdio.h>

#ifdef ANSI_C
  #ifdef C_PLUSPLUS
    #error ANSI_C and C_PLUSPLUS are mutually exclusive
  #else
    int cplusplus_mode = 0;
  #endif
#else
  int cplusplus_mode = 1;
#endif

int complete_file = 0;
int header = 0;
int tabtotab = 8;
int piped = 0;
char * font_size = "11";
char * indentation = "0.5em";
char * comment_font = "itshape";
char * keyword_font = "bfseries";
char * header_font = "slshape";
char * cpp_font = "ttfamily";
char * string_font = "ttfamily";
void substitute(const char *);
void indent(const char *);
void newpage(int);
void usage(const char *);
%}

%%

	if (cplusplus_mode)
	  BEGIN (CPLUSPLUS);

"#"[ \t]*"include"	{ CPP; BEGIN (INCLUDE); }
"#"[ \t]*"define"	|
"#"[ \t]*"pragma"	|
"#"[ \t]*"undef"	|
"#"[ \t]*"if"		|
"#"[ \t]*"ifdef"	|
"#"[ \t]*"ifndef"       |
"#"[ \t]*"elif"		|
"#"[ \t]*"else"		|
"#"[ \t]*"endif"	CPP;

<INCLUDE>"<"[^>]*/">"	{ OUT ("$<${\\"); OUT (string_font);
			  OUT ("{}"); SUB (yytext+1); OUT ("}$>$");
			  input(); BEGIN (INITIAL); }
<INCLUDE>\"[^\"]*/\"	{ OUT ("\"{\\"); OUT (string_font);
			  OUT ("{}"); SUB (yytext+1); OUT ("}\"");
			  input(); BEGIN (INITIAL); }
<INCLUDE>[ \t]+		ECHO;
<INCLUDE>[\n]		OUT ("\\mbox{}\\\\\\relax\n");
<INCLUDE>.		{ yyless (0); BEGIN (INITIAL); }

"auto"			|
"double"		|
"int"			|
"struct"		|
"break"			|
"else"			|
"long"			|
"switch"		|
"case"			|
"enum"			|
"register"		|
"typedef"		|
"char"			|
"extern"		|
"return"		|
"union"			|
"const"			|
"float"			|
"short"			|
"unsigned"		|
"continue"		|
"for"			|
"signed"		|
"void"			|
"default"		|
"goto"			|
"sizeof"		|
"volatile"		|
"do"			|
"if"			|
"static"		|
"while"			|
<CPLUSPLUS>"new"	|
<CPLUSPLUS>"delete"	|
<CPLUSPLUS>"this"	|
<CPLUSPLUS>"operator"	|
<CPLUSPLUS>"class"	|
<CPLUSPLUS>"public"	|
<CPLUSPLUS>"protected"	|
<CPLUSPLUS>"private"	|
<CPLUSPLUS>"virtual"	|
<CPLUSPLUS>"friend"	|
<CPLUSPLUS>"inline"	|
<CPLUSPLUS>"overload"	KEY;


"->"			SYM ("rightarrow");
"<<"			SYM ("ll");
">>"			SYM ("gg");
"<="			OUT ("$<=$");
">="			OUT ("$>=$");
"!="			OUT ("!=");
"||"			SYM ("mid\\mid");
"..."			SYM ("ldots");
"*="			SYM ("ast=");
"<<="			SYM ("ll=");
">>="			SYM ("gg=");
"^="			SYM ("vee=");
"|="			SYM ("mid=");
"~"			SYM ("sim");
"*"			SYM ("ast");
"^"			SYM ("wedge");
"|"			SYM ("mid");
<CPLUSPLUS>"->*"	SYM ("rightarrow\\ast");
"/"			OUT ("$/$");
"<"			OUT ("$<$");
">"			OUT ("$>$");
"&&"			OUT ("\\&\\&");
"%="			OUT ("\\%=");
"&="			OUT ("\\&=");
"{"			OUT ("\\{");
"}"			OUT ("\\}");
"&"			OUT ("\\&");
"%"			OUT ("\\%");
"--"			OUT ("$--$");
"-="                    OUT ("$-=$");
<CPLUSPLUS>".*"		OUT (".$\\ast$");
"-"                     OUT ("$-$");
"?"			|
":"			|
"="			|
","			|
"."			|
";"			|
"!"			|
"+"			|
"/="			|
"=="			|
"++"			|
"+="			|
"("			|
")"			|
"["			|
"]"			|
<CPLUSPLUS>"::"		ECHO;

<CPLUSPLUS>[a-zA-Z_$][a-zA-Z_$0-9]*	|
[a-zA-Z_][a-zA-Z_0-9]*			SUB (yytext);

^"/*"				{ BEGIN (BCOMMENT);
				  OUT ("{$/\\ast$\\");
				  OUT (comment_font); OUT ("{}"); }
"/*"				{ BEGIN (BCOMMENT);
				  if (hfillon==1) { OUT ("\\hfill"); }
				  OUT ("{$/\\ast$\\");
				  OUT (comment_font); OUT ("{}"); }
<BCOMMENT>"*/"			{ BEGIN (INITIAL); OUT ("$\\ast/$}"); }
<BCOMMENT>"\n"			OUT ("\\mbox{}\\\\\\relax\n");
<BCOMMENT>^[ \t]+		IND;
<BCOMMENT>[\201-\237][\100-\176]  |
<BCOMMENT>[\201-\237][\200-\374]  |
<BCOMMENT>[\340-\374][\100-\176]  |
<BCOMMENT>[\340-\374][\200-\374]  SUB (yytext);
<BCOMMENT>"//TeX{"            { BEGIN (TEXINCOMMENT); OUT ("{") ;}
<BCOMMENT>.			SUB (yytext);

"//TeX{"                      { BEGIN (TEXCOMMAND); OUT ("{"); }
<TEXCOMMAND>"//TeX}"\n |
<TEXCOMMAND>"//}"\n           { BEGIN (INITIAL); OUT ("}\\mbox{}\\\\\\relax\n"); }
<TEXCOMMAND>"//TeX}" |
<TEXCOMMAND>"//}"             { BEGIN (INITIAL); OUT ("}");}
<TEXINCOMMENT>"//TeX}"\n |
<TEXINCOMMENT>"//}"\n         { BEGIN (BCOMMENT); OUT ("}\\mbox{}\\\\\\relax\n"); }
<TEXINCOMMENT>"//TeX}" |
<TEXINCOMMENT>"//}"           { BEGIN (BCOMMENT); OUT ("}") ;}
<TEXINCOMMENT>"\\beginfree"   { OUT ("}}\\end{flushleft}{") ; }
<TEXINCOMMENT>"\\endfree"     { OUT ("}\\begin{flushleft}{\\") ;
                                OUT (comment_font) ; OUT ("{") ; }
<TEXCOMMAND,TEXINCOMMENT>"//" { OUT ("//"); }
<TEXCOMMAND,TEXINCOMMENT>"\\" { OUT ("\\"); }
<TEXCOMMAND,TEXINCOMMENT>\n   { OUT ("\n"); }
<TEXCOMMAND,TEXINCOMMENT>[\201-\237][\100-\176]  |
<TEXCOMMAND,TEXINCOMMENT>[\201-\237][\200-\374]  |
<TEXCOMMAND,TEXINCOMMENT>[\340-\374][\100-\176]  |
<TEXCOMMAND,TEXINCOMMENT>[\340-\374][\200-\374]  SUB (yytext);
<TEXCOMMAND,TEXINCOMMENT>.    ECHO;

"//-"                         { BEGIN (CPCOMMENT); OUT ("//{\\"); 
                                OUT (comment_font); OUT ("{}"); }
<CPCOMMENT>\n                 { OUT ("}"); BEGIN (INITIAL);
                                OUT ("\\mbox{}\\\\\\relax\n"); }
<CPCOMMENT>[\201-\237][\100-\176]  |
<CPCOMMENT>[\201-\237][\200-\374]  |
<CPCOMMENT>[\340-\374][\100-\176]  |
<CPCOMMENT>[\340-\374][\200-\374]  SUB (yytext);
<CPCOMMENT>^[ \t]+		   IND;
<CPCOMMENT>.                       SUB (yytext);

L?\" 				{ BEGIN (STRING); OUT ("{\\");
				  OUT (string_font); OUT ("\""); }
<STRING>\\\\			OUT ("$\\backslash\\backslash$");
<STRING>\\[bfnrtv'"]		{ OUT ("$\\backslash$"); SUB (yytext+1); }
<STRING>\"			{ BEGIN (INITIAL); OUT ("\"}"); }
<STRING>"\n"			OUT ("\\mbox{}\\\\\\relax\n");
<STRING>^[ \t]+			IND;
<STRING>[\201-\237][\100-\176]  |
<STRING>[\201-\237][\200-\374]  |
<STRING>[\340-\374][\100-\176]  |
<STRING>[\340-\374][\200-\374]  |
<STRING>.			SUB (yytext);

([0-9]*\.[0-9]+[fFlL]?)		 |
([0-9]+\.[0-9]*[fFlL]?)		 |
([0-9]*\.?[0-9]+[eE][+-]?[0-9]+) |
([0-9]+\.?[0-9]*[eE][+-]?[0-9]+) ECHO;

[0-9]+[uUlL]?			ECHO;

L?'[ -~]'			|
L?'\\[ntvbrfa\\?'"]'		|
L?'\\[0-7]{1,3}'		|
L?'\\x[0-9a-fA-F]{1,2}'		SUB (yytext);

0[0-7]+[uUlL]?			ECHO;

0x[0-9a-fA-F]+[uUlL]?		ECHO;

"\\\n"				OUT ("$\\backslash$\\\\\n");
^[ \t]+				IND;
[ \t]+				ECHO;
"\f"[\n]?			OUT ("\\newpage\n");
"\n"				OUT ("\\mbox{}\\\\\\relax\n");

%%

int iskanji1(unsigned char c)
{
    c &= 0xff;

    if((c>=0x81 && c<=0x9f) || (c>=0xe0 && c<=0xfc)) return 1;
    else return 0;
}

int iskanji2(unsigned char c)
{
    c &= 0xff;

    if((c>=0x40) && (c<=0xfc) && (c!=0x7f)) return 1;
    else return 0;
}

void substitute (const char * input)
{
  while (*input)
    {
      if(iskanji1((unsigned char)(*input)) &&
         iskanji2((unsigned char)(*(input+1)))) {
         printf("%c%c", *input, *(input+1));
         input++;
         input++;
         continue;
      }
      switch (*input)
	{
	case '_':
	case '&':
	case '#':
	case '$':
	case '%':
	case '{':
	case '}':
	  printf ("\\%c", *input);
	  break;
	case '+':
        case '-':
	case '=':
	case '<':
	case '>':
	  printf ("$%c$", *input);
	  break;
	case '*':
	  printf ("$\\ast$");
	  break;
	case '|':
	  printf ("$\\mid$");
	  break;
	case '\\':
	  printf ("$\\backslash$");
	  break;
	case '^':
	  printf ("$\\wedge$");
	  break;
	case '~':
	  printf ("$\\sim$");
	  break;
	default:
	  printf ("%c", *input);
	  break;
	}
      input++;
    }
}

void
indent(const char * blanks)
{
  int i;
  
  i = 0;
  while (*blanks)
    {
      if (*blanks == ' ')
	{
	  i++;
	}
      else
	{
	  while (++i % tabtotab) ;
	}
      blanks++;
    }
  printf ("\\hspace*{%d\\indentation}", i);
}

#include "getopt.h"
#include <string.h>
#include <fcntl.h>
#include <time.h>

extern char * version_string;

static struct option opts[] =
{
  {"ansi-c", 0, 0, 'a'},
  {"c-plusplus", 0, 0, 'p'},
  {"complete-file", 0, 0, 'c'},
  {"font-size", 1, 0, 's'},
  {"indentation", 1, 0, 'i'},
  {"header", 0, 0, 'h'},
  {"piped", 0, 0, 't'},
  {"output", 1, 0, 'o'},
  {"tabstop", 1, 0, 'T'},
  {"comment-font", 1, 0, 'C'},
  {"string-font", 1, 0, 'S'},
  {"keyword-font", 1, 0, 'K'},
  {"header-font", 1, 0, 'H'},
  {"cpp-font", 1, 0, 'P'},
  {"version", 0, 0, 'V'},
  {0, 0, 0, 0}
};

#ifdef WIN32___
#include <ctype.h>
#endif

main (int argc, char** argv)
{
  int c;
  int index;
  int i;
  int has_filename;
  char * input_name;
  char * output_name;
  char * program_name;
  time_t time_value;
  char *time_string;
  char *akp;
  
  (void)time(&time_value);
  time_string = ctime(&time_value);
  if(akp = strrchr(time_string,'\n')) *akp = '\0';

#ifdef WIN32___
  akp = argv[0];
  while(*akp) {
    if(*akp == '\\') *akp = '/';
    *akp = tolower(*akp);
    akp++;
  }
  if(akp = strrchr(argv[0],'/')) {
    akp++;
    argv[0] = akp;
  }
  if(akp = strrchr(argv[0],'.'))
    *akp = '\0';

  (void)setmode(fileno(stdout), O_BINARY);
#endif

  input_name = "Standard Input";
  output_name = 0;

  program_name = strrchr (argv[0], '/');
  if (program_name == NULL)
    {
      program_name = argv[0];
    }
  else
    {
      program_name++;
    }
  
#ifdef USE_NAME 
#if defined(ANSI_C) || defined(C_PLUSPLUS)
#error USE_NAME, ANSI_C and C_PLUSPLUS are mutually exclusive
#else
  cplusplus_mode = (strchr (program_name, '+') != 0);
#endif
#endif
  
  while ((c = getopt_long (argc, argv,
			   "acpo:s:i:thT:C:H:S:K:P:V", opts, &index))
	 != EOF)
    {
      if (c == 0)
	{
	  c = opts[index].val;
	}
      switch (c)
	{
	case 'a':
	  cplusplus_mode = 0;
	  break;
	case 'p':
	  cplusplus_mode = 1;
	  break;
	case 'c':
	  complete_file = 1;
	  break;
	case 'o':
	  if (piped)
	    {
	      fprintf (stderr,
		       "%s: Can't use {-t,+pipe} and {-o,+output} together\n",
		       program_name);
	      exit(5);
	    }
	  output_name = optarg;
	  break;
	case 's':
	  font_size = optarg;
	  break;
	case 'i':
	  indentation = optarg;
	  break;
	case 'T':
	  tabtotab = atoi(optarg);
	  break;
	    case 't':
	      if (output_name != 0)
	    {
	      fprintf (stderr,
		       "%s: Can't use {-t,+pipe} and {-o,+output} together\n",
		       program_name);
	      exit(5);
	    }
	  piped = 1;
	  break;
	case 'h':
	  header = 1;
	  complete_file = 1;
	  break;
	case 'C':
	  comment_font = optarg;
	  break;
	case 'H':
	  header_font = optarg;
	  break;
	case 'P':
	  cpp_font = optarg;
	  break;
	case 'S':
	  string_font = optarg;
	  break;
	case 'K':
	  keyword_font = optarg;
	  break;
	case 'V':
	  fprintf (stderr, "%s\n", version_string);
	  break;
	default:
	  usage(program_name);
	}
    }
  has_filename = (argc - optind == 1);
  if (has_filename)
    {
      input_name = argv[optind];
      if (freopen (input_name, "r", stdin) == NULL)
	{
	  fprintf (stderr, "%s: Can't open `%s' for reading\n",
		   program_name, input_name);
	  exit (2);
	}
    }
  if ((output_name == 0) && !piped)
    {
      char * tmp;
      if (has_filename)
	{
	  char * point;
	  
	  point = strrchr (input_name, '/');
	  if (point == 0)
	    {
	      point = input_name;
	    }
	  else
	    {
	      point++;
	    }
	  tmp = (char*)malloc (strlen (point) + 1);
	  if (tmp == 0)
	    {
	      fprintf (stderr, "%s: Virtual memory exhausted\n", program_name);
	      exit (3);
	    }
	  strcpy (tmp, point);
	  point = strrchr (tmp, '.');
	  if (point != 0)
	    {
	      *point = '\0';
	    }
	}
      else
	{
	  tmp = program_name;
	}
      output_name = (char *)malloc (strlen (tmp) + 4);
      if (output_name == 0)
	{
	  fprintf (stderr, "%s: Virtual memory exhausted\n", program_name);
	  exit (3);
	}
      strcpy (output_name, tmp);
      strcat (output_name, ".tex");
    }
  if (!piped)
    {
#ifdef WIN32___
      if (freopen (output_name, "wb", stdout) == NULL)
#else
      if (freopen (output_name, "w", stdout) == NULL)
#endif
	{
	  fprintf (stderr, "%s: Can't open `%s' for writing\n",
		   program_name, output_name);
	  exit (4);
	}
    }
  printf ("\
%%#! platex\n\
%% This file was automatically produced at %s by\n\
%% %s", time_string, program_name);
  for (i = 1; i < argc; i++)
    {
      printf(" %s", argv[i]);
    }
  if (!has_filename)
    {
      printf(" (from Standard Input)");
    }
  printf("\n%%\n");
  if (complete_file)
    {
      if (header)
	{
	  if (strcmp (font_size, "10") == 0)
	    {
	      printf ("\\documentclass{jarticle}\n");
              printf ("\\usepackage{fancyhdr}\n");
	    }
	  else
	    {
	      printf ("\\documentclass[%spt]{jarticle}\n",
		      font_size);
              printf ("\\usepackage{fancyhdr}\n");
	    }
	}
      else
	{
	  if (strcmp (font_size, "10") == 0)
	    {
	      printf ("\\documentclass{jarticle}\n");
	    }
	  else
	    {
	      printf ("\\documentclass[%spt]{jarticle}\n", font_size);
	    }
	}
      printf ("\\setlength{\\textwidth}{15.5cm}\n");
      printf ("\\setlength{\\textheight}{22.5cm}\n");
      printf ("\\setlength{\\hoffset}{-1.2cm}\n");
      printf ("\\setlength{\\voffset}{-1.2cm}\n");
      if (header)
	{
	  printf ("\\chead{\\%s Produced from ", header_font);
	  substitute(input_name);
	  printf (" at %s}\n", time_string);
	  printf ("\\cfoot{\\rmfamily\\thepage}\n");
	  printf ("\\addtolength{\\headheight}{14pt}\n");
	  printf ("\\pagestyle{fancy}\n");
	}
      printf ("\\begin{document}\n");
    }
  printf ("\\expandafter\\ifx\\csname indentation\\endcsname\\relax%\n");
  printf ("\\newlength{\\indentation}\\fi\n");
  printf ("\\setlength{\\indentation}{%s}\n", indentation);
  printf ("\\begin{flushleft}\n");
  yylex();
  printf ("\\end{flushleft}\n");
  if (complete_file)
    {
      printf ("\\end{document}\n");
    }
  exit (0);
}

void
usage(const char * name)
{
  fprintf (stderr, "%s\n", version_string);
  fprintf (stderr, "\
Usage: %s [-a] [-c] [-h] [-i length] [-o path] [-p] [-s size] [-t]\n\
       [-C font][-H font] [-K font] [-P font] [-S font] [-T wide] [-V]\n\
       [+ansi-c] [+complete-file] [+header] [+indentation length]\n\
       [+output path] [+c-plusplus] [+font-size size] [+pipe]\n\
       [+comment-font font] [+keyword-font font] [+cpp-font font]\n\
       [+header-font font] [+string-font font] [+tabstop wide]\n\
       [+version] [path]\n", name);
  exit (1);
}

yywrap(){}
