/*
 * Copyright (C) 2007 Keith Stribley
 * 
 * Distributable under the terms of either the Common Public License or the
 * GNU Lesser General Public License, as specified in the LICENSING.txt file.
 *
 */
package org.sil.graphite.ui;

import java.awt.Color;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Insets;
import java.awt.LinearGradientPaint;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.font.FontRenderContext;
import java.awt.font.GlyphVector;

import javax.swing.AbstractButton;
import javax.swing.Icon;
import javax.swing.JComponent;
import javax.swing.JMenuItem;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;
import javax.swing.plaf.ButtonUI;
import javax.swing.plaf.ColorUIResource;
import javax.swing.plaf.basic.BasicButtonUI;
import javax.swing.plaf.basic.BasicGraphicsUtils;
import javax.swing.plaf.multi.MultiButtonUI;

import org.sil.graphite.awt.GraphiteAwtFont;
import org.sil.graphite.awt.GraphiteFontManager;
import org.sil.graphite.swing.PaintHelper;

/**
 * @author keith
 *
 */
public class GrButtonUI extends ButtonUI
{
    double graphiteWidth = Double.NaN;
    boolean callBack = false;
    
    public GrButtonUI()
    {
        
    }
    
    /* (non-Javadoc)
     * @see javax.swing.plaf.ComponentUI#paint(java.awt.Graphics, javax.swing.JComponent)
     */
    @Override
    public void paint(Graphics g, JComponent c)
    {
        PaintHelper.preparePaint(c, g);
        if (callBack == false && c instanceof AbstractButton)
        {
            AbstractButton b = (AbstractButton)c;
            BasicButtonUI bbui = null;
            if (b.getUI() instanceof MultiButtonUI)
            {
                callBack = true;
                MultiButtonUI mbui = (MultiButtonUI)b.getUI();
                //String oldText = b.getText();
                Color foreground = b.getForeground();
                //b.setForeground(b.getBackground());
                if (mbui.getUIs()[0] instanceof BasicButtonUI)
                {
                    bbui = (BasicButtonUI)mbui.getUIs()[0];
                }
                //mbui.getUIs()[0].paint(g, c);
                //b.setForeground(foreground);
                //b.setText(oldText);
                callBack = false;
            }
            
            
            if (GraphiteFontManager.isGraphite(b.getFont()))
            {
              GraphiteAwtFont font = GraphiteFontManager.getFont(b.getFont());
              if (font == null) return;
              Graphics2D g2d = (Graphics2D)g;
              FontRenderContext frc = g2d.getFontRenderContext();
              GlyphVector gv = font.layoutGlyphVector(frc, b.getText().toCharArray(), 
                  0, b.getText().length(), Font.LAYOUT_LEFT_TO_RIGHT);
              FontMetrics fm = b.getFontMetrics(font);
              graphiteWidth = gv.getGlyphPosition(gv.getNumGlyphs()).getX();
              // find the text position
              Insets i = b.getInsets();
              
              Rectangle viewRect = new Rectangle(i.left,i.top,
                      b.getWidth()- (i.right + i.left),
                      b.getHeight() - (i.bottom + i.top));
              
              Rectangle textRect = new Rectangle(0,0,0,0);
              Rectangle iconRect = new Rectangle(0,0,0,0);
              SwingUtilities.layoutCompoundLabel(
                      b, fm, b.getText(), b.getIcon(), 
                      b.getVerticalAlignment(), b.getHorizontalAlignment(),
                      b.getVerticalTextPosition(), b.getHorizontalTextPosition(),
                      viewRect, iconRect, textRect, 
                      b.getText() == null ? 0 : b.getIconTextGap());
              // overwrite background
              g2d.setColor(b.getBackground());
//              if(b instanceof JMenuItem)
//              {
//                  g2d.fill(new Rectangle(0, 0, b.getWidth(), b.getHeight()));
//              }
              if (UIManager.getLookAndFeel().getName().equals("Metal"))
              {
                  String gradientName = "Button.gradient";
                  float [] fractions = { 0f, 0.3f, 1f };
                  java.util.List gradient = (java.util.List)UIManager.get(gradientName);
                  Color [] colors = { ((ColorUIResource)gradient.get(2)),
                          ((ColorUIResource)gradient.get(3)),
                          ((ColorUIResource)gradient.get(4))
                  };
                  if (b instanceof JMenuItem)
                  {
                      gradientName = "MenuBar.gradient";
                      float [] menuFractions = { 0f, 1f };
                      gradient = (java.util.List)UIManager.get(gradientName);
                      Color [] menuColors = { ((ColorUIResource)gradient.get(2)),
                              ((ColorUIResource)gradient.get(3))
                      };
                      fractions = menuFractions;
                      colors = menuColors;
                  }
                  LinearGradientPaint lgp = new LinearGradientPaint(0f, 0f,
                          0f, (float)b.getHeight(), fractions, colors);
                  g2d.setPaint(lgp);
                  
                  g2d.fill(new Rectangle(0, 0, b.getWidth(), b.getHeight()));
                  // redraw the icon
                  Icon icon = b.getIcon();
                  Icon tmpIcon = null;
                  if (b.isSelected()) 
                  {
                      tmpIcon = b.getSelectedIcon();
                      if (!b.isEnabled()) 
                          tmpIcon = b.getDisabledSelectedIcon();
                      else if (b.isRolloverEnabled())
                          tmpIcon = b.getRolloverSelectedIcon();
                  }
                  else if (!b.isEnabled()) 
                  {
                      tmpIcon = b.getDisabledIcon();
                  }
                  else if (b.isRolloverEnabled())
                  {
                      tmpIcon = b.getRolloverIcon();
                  }
                  if (tmpIcon != null) icon = tmpIcon;
                  if (icon != null)
                      icon.paintIcon(c, g, iconRect.x, iconRect.y);
              }              
              else
              {
                  g2d.fill(textRect);
              }

              // correct position for graphite width
              double xOffset = textRect.x + (textRect.width - graphiteWidth)/2.0; 
              if (xOffset < 0) xOffset = 0;
              g2d.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, 
                      RenderingHints.VALUE_TEXT_ANTIALIAS_LCD_HRGB);
              if (b.getModel().isEnabled())
              {
                  g2d.setColor(b.getForeground());
                  g2d.drawGlyphVector(gv, (float) (xOffset),
                          textRect.y + fm.getAscent() + fm.getLeading());
              }
              else
              {
                  g2d.setColor(b.getBackground().brighter());
                  g2d.drawGlyphVector(gv, (float) (xOffset),
                      textRect.y + fm.getAscent() + fm.getLeading());
                  
                  g2d.setColor(b.getBackground().darker());
                  g2d.drawGlyphVector(gv, (float) (xOffset - 1),
                          textRect.y + fm.getAscent() + fm.getLeading() - 1);
              }
              return;
            }
        }
    }

    /* (non-Javadoc)
     * @see javax.swing.plaf.ComponentUI#update(java.awt.Graphics, javax.swing.JComponent)
     */
    @Override
    public void update(Graphics g, JComponent c)
    {
        // must do nothing to get default rendering
        
        paint(g, c);
    }

    /* (non-Javadoc)
     * @see javax.swing.plaf.ComponentUI#installUI(javax.swing.JComponent)
     */
    @Override
    public void installUI(JComponent c)
    {
        // TODO Auto-generated method stub
        //super.installUI(c);
    }

    /* (non-Javadoc)
     * @see javax.swing.plaf.ComponentUI#uninstallUI(javax.swing.JComponent)
     */
    @Override
    public void uninstallUI(JComponent c)
    {
        // TODO Auto-generated method stub
       // super.uninstallUI(c);
    }

}
