/*
 * Copyright (C) 2007 Keith Stribley
 * 
 * Distributable under the terms of either the Common Public License or the
 * GNU Lesser General Public License, as specified in the LICENSING.txt file.
 *
 */
package org.sil.graphite.ui;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Insets;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.font.FontRenderContext;
import java.awt.font.GlyphVector;

import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.SwingConstants;
import javax.swing.SwingUtilities;
import javax.swing.plaf.ComponentUI;
import javax.swing.plaf.LabelUI;
import javax.swing.plaf.basic.BasicLabelUI;
import javax.swing.plaf.multi.MultiLabelUI;

import org.sil.graphite.awt.GraphiteAwtFont;
import org.sil.graphite.awt.GraphiteFontManager;

/**
 * Graphite enabled Labels
 * TODO disabled text, pneumonics, elipsis... 
 */
public class GrLabelUI extends LabelUI
{
    double graphiteWidth = 0;
    double graphiteHeight = 0;
    //boolean callBack = false;
    
    public GrLabelUI()
    {
        
    }
    
    /* (non-Javadoc)
     * @see javax.swing.plaf.ComponentUI#paint(java.awt.Graphics, javax.swing.JComponent)
     */
    @Override
    public void paint(Graphics g, JComponent c)
    {
        if (c instanceof JLabel)
        {
            // TODO support pneumonics and disabled text support
            JLabel l = (JLabel)c;
            if (GraphiteFontManager.isGraphite(l.getFont()))
            {
                Graphics2D g2d = (Graphics2D)g;
                
                GraphiteAwtFont font = GraphiteFontManager.getFont(l.getFont());
                if (font == null) return;
                // not sure whether this should be hard coded, but at least
                // it gives better than the default
                g2d.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, 
                        RenderingHints.VALUE_TEXT_ANTIALIAS_LCD_HRGB);
                FontMetrics fm = l.getFontMetrics(font);
                // where is the text drawn?
                Insets i = l.getInsets();
                Rectangle viewRect = new Rectangle(i.left,i.top,
                        l.getWidth()- (i.right + i.left),
                        l.getHeight() - (i.bottom + i.top));
                
                Rectangle textRect = new Rectangle(0,0,0,0);
                Rectangle iconRect = new Rectangle(0,0,0,0);
                SwingUtilities.layoutCompoundLabel(
                        l, fm, l.getText(), l.getIcon(), 
                        l.getVerticalAlignment(), l.getHorizontalAlignment(),
                        l.getVerticalTextPosition(), l.getHorizontalTextPosition(),
                        viewRect, iconRect, textRect,
                        l.getText() == null ? 0 : l.getIconTextGap());
                
                // this is rather a hack, but paint over the original with the background
                g.setColor(c.getBackground());
                g2d.fill(textRect);
                
                
                FontRenderContext frc = g2d.getFontRenderContext();
                GlyphVector gv = font.layoutGlyphVector(frc, l.getText().toCharArray(), 
                  0, l.getText().length(), Font.LAYOUT_LEFT_TO_RIGHT);
                graphiteHeight = fm.getHeight();
                graphiteWidth = gv.getGlyphPosition(gv.getNumGlyphs()).getX();
                float xOffset = 0;
                if (l.getHorizontalAlignment() == SwingConstants.RIGHT)
                {
                    xOffset = (float)(l.getWidth() - l.getInsets().left - l.getInsets().right - graphiteWidth);
                }
                else if (l.getHorizontalAlignment() == SwingConstants.CENTER)
                {
                    xOffset = 0.5f * (float)(l.getWidth() - graphiteWidth - l.getInsets().left - l.getInsets().right);
                }
                if (l.isEnabled())
                    g.setColor(c.getForeground());
                else
                    g.setColor(c.getBackground().brighter());
                g2d.drawGlyphVector(gv, textRect.x + xOffset, 
                    textRect.y + fm.getAscent() + fm.getLeading());
                if (l.isEnabled() == false)
                {
                    g.setColor(c.getBackground().darker());
                    g2d.drawGlyphVector(gv, textRect.x + xOffset - 1, 
                        textRect.y + fm.getAscent() + fm.getLeading() - 1);
                }
                getPreferredSize(l);
                return;
            }
        }
        super.paint(g, c);
    }

    /* (non-Javadoc)
     * @see javax.swing.plaf.ComponentUI#update(java.awt.Graphics, javax.swing.JComponent)
     */
    @Override
    public void update(Graphics g, JComponent c)
    {
        // call super, which then calls paint
        //super.update(g, c);
        paint(g, c);
    }

    /* (non-Javadoc)
     * @see javax.swing.plaf.ComponentUI#installUI(javax.swing.JComponent)
     */
    @Override
    public void installUI(JComponent c)
    {
        
        super.installUI(c);
    }

    /* (non-Javadoc)
     * @see javax.swing.plaf.ComponentUI#uninstallUI(javax.swing.JComponent)
     */
    @Override
    public void uninstallUI(JComponent c)
    {
        
        super.uninstallUI(c);
    }

//    /* (non-Javadoc)
//     * @see javax.swing.plaf.ComponentUI#getPreferredSize(javax.swing.JComponent)
//     */
//    @Override
//    public Dimension getPreferredSize(JComponent c)
//    {
//        if (callBack == false && graphiteWidth > 0)
//        {
//            // fix up the width
//            Dimension dim = new Dimension();
//            java.awt.Insets insets = c.getInsets();
//            dim.setSize(graphiteWidth + insets.left + insets.right, 
//                graphiteHeight + insets.top + insets.bottom);
//            callBack = true;
//            c.setPreferredSize(dim);
//            callBack = false;
//            return dim;
//        }
//        return null;
//    }

    LabelUI getPrimaryUI(JComponent c)
    {
        if (c instanceof JLabel)
        {
            JLabel l = (JLabel)c;
            if (l.getUI() instanceof MultiLabelUI)
            {
                MultiLabelUI labelUI = (MultiLabelUI)l.getUI();
                return (LabelUI)labelUI.getUIs()[0];
            }
            else return l.getUI();
        }
        return this;
    }
}
